@@ -25,8 +25,11 @@ import (
 	"net/http"
 	"os"
 	"path/filepath"
+	"runtime"
 	"time"
 
+	"kubevirt.io/kubevirt/pkg/flavor"
+
 	"github.com/emicklei/go-restful"
 	vsv1beta1 "github.com/kubernetes-csi/external-snapshotter/v2/pkg/apis/volumesnapshot/v1beta1"
 	"github.com/prometheus/client_golang/prometheus"
@@ -40,16 +43,26 @@ import (
 	"k8s.io/client-go/tools/leaderelection"
 	"k8s.io/client-go/tools/leaderelection/resourcelock"
 	"k8s.io/client-go/tools/record"
+	"k8s.io/client-go/util/flowcontrol"
+
+	"kubevirt.io/kubevirt/pkg/util/ratelimiter"
+
+	"kubevirt.io/kubevirt/pkg/virt-controller/watch/topology"
 
 	"kubevirt.io/kubevirt/pkg/healthz"
+	"kubevirt.io/kubevirt/pkg/monitoring/profiler"
 
-	snapshotv1 "kubevirt.io/client-go/apis/snapshot/v1alpha1"
+	poolv1 "kubevirt.io/api/pool/v1alpha1"
+	snapshotv1 "kubevirt.io/api/snapshot/v1alpha1"
 	"kubevirt.io/client-go/kubecli"
 	"kubevirt.io/client-go/log"
 	clientutil "kubevirt.io/client-go/util"
 	"kubevirt.io/kubevirt/pkg/certificates/bootstrap"
 	containerdisk "kubevirt.io/kubevirt/pkg/container-disk"
 	"kubevirt.io/kubevirt/pkg/controller"
+
+	"kubevirt.io/kubevirt/pkg/monitoring/perfscale"
+	vmiprom "kubevirt.io/kubevirt/pkg/monitoring/vmistats" // import for prometheus metrics
 	"kubevirt.io/kubevirt/pkg/service"
 	"kubevirt.io/kubevirt/pkg/util"
 	"kubevirt.io/kubevirt/pkg/util/webhooks"
@@ -59,6 +72,7 @@ import (
 	"kubevirt.io/kubevirt/pkg/virt-controller/watch/drain/disruptionbudget"
 	"kubevirt.io/kubevirt/pkg/virt-controller/watch/drain/evacuation"
 	"kubevirt.io/kubevirt/pkg/virt-controller/watch/snapshot"
+	workloadupdater "kubevirt.io/kubevirt/pkg/virt-controller/watch/workload-updater"
 )
 
 const (
@@ -66,7 +80,8 @@ const (
 
 	defaultHost = "0.0.0.0"
 
-	launcherImage = "virt-launcher"
+	launcherImage       = "virt-launcher"
+	launcherQemuTimeout = 240
 
 	imagePullSecret = ""
 
@@ -74,10 +89,13 @@ const (
 
 	ephemeralDiskDir = virtShareDir + "-ephemeral-disks"
 
-	defaultControllerThreads = 3
+	defaultControllerThreads         = 3
+	defaultSnapshotControllerThreads = 6
+	defaultVMIControllerThreads      = 10
 
 	defaultLauncherSubGid                 = 107
 	defaultSnapshotControllerResyncPeriod = 5 * time.Minute
+	defaultNodeTopologyUpdatePeriod       = 30 * time.Second
 
 	defaultPromCertFilePath = "/etc/virt-controller/certificates/tls.crt"
 	defaultPromKeyFilePath  = "/etc/virt-controller/certificates/tls.key"
@@ -89,17 +107,19 @@ var (
 
 	leaderGauge = prometheus.NewGauge(
 		prometheus.GaugeOpts{
-			Name: "leading_virt_controller",
+			Name: "kubevirt_virt_controller_leading",
 			Help: "Indication for an operating virt-controller.",
 		},
 	)
 
 	readyGauge = prometheus.NewGauge(
 		prometheus.GaugeOpts{
-			Name: "ready_virt_controller",
+			Name: "kubevirt_virt_controller_ready",
 			Help: "Indication for a virt-controller that is ready to take the lead.",
 		},
 	)
+
+	apiHealthVersion = new(healthz.KubeApiHealthzVersion)
 )
 
 type VirtControllerApp struct {
@@ -123,20 +143,31 @@ type VirtControllerApp struct {
 
 	clusterConfig *virtconfig.ClusterConfig
 
+	pdbInformer cache.SharedIndexInformer
+
 	persistentVolumeClaimCache    cache.Store
 	persistentVolumeClaimInformer cache.SharedIndexInformer
 
 	rsController *VMIReplicaSet
 	rsInformer   cache.SharedIndexInformer
 
+	poolController *PoolController
+	poolInformer   cache.SharedIndexInformer
+
 	vmController *VMController
 	vmInformer   cache.SharedIndexInformer
 
+	controllerRevisionInformer cache.SharedIndexInformer
+
 	dataVolumeInformer cache.SharedIndexInformer
+	cdiInformer        cache.SharedIndexInformer
+	cdiConfigInformer  cache.SharedIndexInformer
 
 	migrationController *MigrationController
 	migrationInformer   cache.SharedIndexInformer
 
+	workloadUpdateController *workloadupdater.WorkloadUpdateController
+
 	snapshotController        *snapshot.VMSnapshotController
 	restoreController         *snapshot.VMRestoreController
 	vmSnapshotInformer        cache.SharedIndexInformer
@@ -147,9 +178,15 @@ type VirtControllerApp struct {
 
 	crdInformer cache.SharedIndexInformer
 
+	flavorInformer        cache.SharedIndexInformer
+	clusterFlavorInformer cache.SharedIndexInformer
+
+	migrationPolicyInformer cache.SharedIndexInformer
+
 	LeaderElection leaderelectionconfig.Configuration
 
 	launcherImage              string
+	launcherQemuTimeout        int
 	imagePullSecret            string
 	virtShareDir               string
 	virtLibDir                 string
@@ -158,6 +195,7 @@ type VirtControllerApp struct {
 	hotplugDiskDir             string
 	readyChan                  chan bool
 	kubevirtNamespace          string
+	host                       string
 	evacuationController       *evacuation.EvacuationController
 	disruptionBudgetController *disruptionbudget.DisruptionBudgetController
 
@@ -172,6 +210,7 @@ type VirtControllerApp struct {
 	nodeControllerThreads             int
 	vmiControllerThreads              int
 	rsControllerThreads               int
+	poolControllerThreads             int
 	vmControllerThreads               int
 	migrationControllerThreads        int
 	evacuationControllerThreads       int
@@ -181,9 +220,13 @@ type VirtControllerApp struct {
 	restoreControllerThreads          int
 	snapshotControllerResyncPeriod    time.Duration
 
-	caConfigMapName  string
-	promCertFilePath string
-	promKeyFilePath  string
+	caConfigMapName          string
+	promCertFilePath         string
+	promKeyFilePath          string
+	nodeTopologyUpdater      topology.NodeTopologyUpdater
+	nodeTopologyUpdatePeriod time.Duration
+	reloadableRateLimiter    *ratelimiter.ReloadableRateLimiter
+	leaderElector            *leaderelection.LeaderElector
 }
 
 var _ service.Service = &VirtControllerApp{}
@@ -191,6 +234,7 @@ var _ service.Service = &VirtControllerApp{}
 func init() {
 	vsv1beta1.AddToScheme(scheme.Scheme)
 	snapshotv1.AddToScheme(scheme.Scheme)
+	poolv1.AddToScheme(scheme.Scheme)
 
 	prometheus.MustRegister(leaderGauge)
 	prometheus.MustRegister(readyGauge)
@@ -208,8 +252,13 @@ func Execute() {
 
 	log.InitializeLogging("virt-controller")
 
-	app.clientSet, err = kubecli.GetKubevirtClient()
-
+	app.reloadableRateLimiter = ratelimiter.NewReloadableRateLimiter(flowcontrol.NewTokenBucketRateLimiter(virtconfig.DefaultVirtControllerQPS, virtconfig.DefaultVirtControllerBurst))
+	clientConfig, err := kubecli.GetKubevirtClientConfig()
+	if err != nil {
+		panic(err)
+	}
+	clientConfig.RateLimiter = app.reloadableRateLimiter
+	app.clientSet, err = kubecli.GetKubevirtClientFromRESTConfig(clientConfig)
 	if err != nil {
 		golog.Fatal(err)
 	}
@@ -222,28 +271,46 @@ func Execute() {
 		golog.Fatalf("Error searching for namespace: %v", err)
 	}
 
+	host, err := os.Hostname()
+	if err != nil {
+		golog.Fatalf("unable to get hostname: %v", err)
+	}
+	app.host = host
+
 	ctx, cancel := context.WithCancel(context.Background())
 	stopChan := ctx.Done()
 	app.ctx = ctx
 
 	app.informerFactory = controller.NewKubeInformerFactory(app.restClient, app.clientSet, nil, app.kubevirtNamespace)
 
-	configMapInformer := app.informerFactory.ConfigMap()
 	app.crdInformer = app.informerFactory.CRD()
 	app.kubeVirtInformer = app.informerFactory.KubeVirt()
 	app.informerFactory.Start(stopChan)
 
-	cache.WaitForCacheSync(stopChan, configMapInformer.HasSynced, app.crdInformer.HasSynced, app.kubeVirtInformer.HasSynced)
-	app.clusterConfig = virtconfig.NewClusterConfig(configMapInformer, app.crdInformer, app.kubeVirtInformer, app.kubevirtNamespace)
+	app.kubeVirtInformer.SetWatchErrorHandler(func(r *cache.Reflector, err error) {
+		apiHealthVersion.Clear()
+		cache.DefaultWatchErrorHandler(r, err)
+	})
+
+	cache.WaitForCacheSync(stopChan, app.crdInformer.HasSynced, app.kubeVirtInformer.HasSynced)
+	app.clusterConfig = virtconfig.NewClusterConfig(app.crdInformer, app.kubeVirtInformer, app.kubevirtNamespace)
 
 	app.reInitChan = make(chan string, 10)
 	app.hasCDI = app.clusterConfig.HasDataVolumeAPI()
 	app.clusterConfig.SetConfigModifiedCallback(app.configModificationCallback)
+	app.clusterConfig.SetConfigModifiedCallback(app.shouldChangeLogVerbosity)
+	app.clusterConfig.SetConfigModifiedCallback(app.shouldChangeRateLimiter)
 
 	webService := new(restful.WebService)
 	webService.Path("/").Consumes(restful.MIME_JSON).Produces(restful.MIME_JSON)
-	webService.Route(webService.GET("/healthz").To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig)).Doc("Health endpoint"))
+	webService.Route(webService.GET("/healthz").To(healthz.KubeConnectionHealthzFuncFactory(app.clusterConfig, apiHealthVersion)).Doc("Health endpoint"))
 	webService.Route(webService.GET("/leader").To(app.leaderProbe).Doc("Leader endpoint"))
+
+	componentProfiler := profiler.NewProfileManager(app.clusterConfig)
+	webService.Route(webService.GET("/start-profiler").To(componentProfiler.HandleStartProfiler).Doc("start profiler endpoint"))
+	webService.Route(webService.GET("/stop-profiler").To(componentProfiler.HandleStopProfiler).Doc("stop profiler endpoint"))
+	webService.Route(webService.GET("/dump-profiler").To(componentProfiler.HandleDumpProfiler).Doc("dump profiler results endpoint"))
+
 	restful.Add(webService)
 
 	app.vmiInformer = app.informerFactory.VMI()
@@ -251,19 +318,22 @@ func Execute() {
 	app.nodeInformer = app.informerFactory.KubeVirtNode()
 
 	app.vmiCache = app.vmiInformer.GetStore()
-	app.vmiRecorder = app.getNewRecorder(k8sv1.NamespaceAll, "virtualmachine-controller")
+	app.vmiRecorder = app.newRecorder(k8sv1.NamespaceAll, "virtualmachine-controller")
 
 	app.rsInformer = app.informerFactory.VMIReplicaSet()
+	app.poolInformer = app.informerFactory.VMPool()
 
 	app.persistentVolumeClaimInformer = app.informerFactory.PersistentVolumeClaim()
 	app.persistentVolumeClaimCache = app.persistentVolumeClaimInformer.GetStore()
 
-	app.informerFactory.K8SInformerFactory().Policy().V1beta1().PodDisruptionBudgets().Informer()
+	app.pdbInformer = app.informerFactory.K8SInformerFactory().Policy().V1beta1().PodDisruptionBudgets().Informer()
 
 	app.vmInformer = app.informerFactory.VirtualMachine()
 
 	app.migrationInformer = app.informerFactory.VirtualMachineInstanceMigration()
 
+	app.controllerRevisionInformer = app.informerFactory.ControllerRevision()
+
 	app.vmSnapshotInformer = app.informerFactory.VirtualMachineSnapshot()
 	app.vmSnapshotContentInformer = app.informerFactory.VirtualMachineSnapshotContent()
 	app.vmRestoreInformer = app.informerFactory.VirtualMachineRestore()
@@ -272,28 +342,37 @@ func Execute() {
 
 	if app.hasCDI {
 		app.dataVolumeInformer = app.informerFactory.DataVolume()
+		app.cdiInformer = app.informerFactory.CDI()
+		app.cdiConfigInformer = app.informerFactory.CDIConfig()
 		log.Log.Infof("CDI detected, DataVolume integration enabled")
 	} else {
 		// Add a dummy DataVolume informer in the event datavolume support
 		// is disabled. This lets the controller continue to work without
 		// requiring a separate branching code path.
 		app.dataVolumeInformer = app.informerFactory.DummyDataVolume()
+		app.cdiInformer = app.informerFactory.DummyCDI()
+		app.cdiConfigInformer = app.informerFactory.DummyCDIConfig()
 		log.Log.Infof("CDI not detected, DataVolume integration disabled")
 	}
 
+	app.flavorInformer = app.informerFactory.VirtualMachineFlavor()
+	app.clusterFlavorInformer = app.informerFactory.VirtualMachineClusterFlavor()
+
+	app.migrationPolicyInformer = app.informerFactory.MigrationPolicy()
+
 	app.initCommon()
 	app.initReplicaSet()
+	app.initPool()
 	app.initVirtualMachines()
 	app.initDisruptionBudgetController()
 	app.initEvacuationController()
 	app.initSnapshotController()
 	app.initRestoreController()
+	app.initWorkloadUpdaterController()
 	go app.Run()
 
-	select {
-	case <-app.reInitChan:
-		cancel()
-	}
+	<-app.reInitChan
+	cancel()
 }
 
 // Detects if a config has been applied that requires
@@ -310,6 +389,22 @@ func (vca *VirtControllerApp) configModificationCallback() {
 	}
 }
 
+// Update virt-controller rate limiter
+func (app *VirtControllerApp) shouldChangeRateLimiter() {
+	config := app.clusterConfig.GetConfig()
+	qps := config.ControllerConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.QPS
+	burst := config.ControllerConfiguration.RestClient.RateLimiter.TokenBucketRateLimiter.Burst
+	app.reloadableRateLimiter.Set(flowcontrol.NewTokenBucketRateLimiter(qps, burst))
+	log.Log.V(2).Infof("setting rate limiter to %v QPS and %v Burst", qps, burst)
+}
+
+// Update virt-controller log verbosity on relevant config changes
+func (vca *VirtControllerApp) shouldChangeLogVerbosity() {
+	verbosity := vca.clusterConfig.GetVirtControllerVerbosity(vca.host)
+	log.Log.SetVerbosityLevel(int(verbosity))
+	log.Log.V(2).Infof("set log verbosity to %d", verbosity)
+}
+
 func (vca *VirtControllerApp) Run() {
 	logger := log.Log
 
@@ -331,45 +426,12 @@ func (vca *VirtControllerApp) Run() {
 		}
 	}()
 
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, leaderelectionconfig.DefaultEndpointName)
-
-	id, err := os.Hostname()
-	if err != nil {
-		golog.Fatalf("unable to get hostname: %v", err)
-	}
-
-	rl, err := resourcelock.New(vca.LeaderElection.ResourceLock,
-		vca.kubevirtNamespace,
-		leaderelectionconfig.DefaultEndpointName,
-		vca.clientSet.CoreV1(),
-		vca.clientSet.CoordinationV1(),
-		resourcelock.ResourceLockConfig{
-			Identity:      id,
-			EventRecorder: recorder,
-		})
-	if err != nil {
-		golog.Fatal(err)
-	}
-
-	leaderElector, err := leaderelection.NewLeaderElector(
-		leaderelection.LeaderElectionConfig{
-			Lock:          rl,
-			LeaseDuration: vca.LeaderElection.LeaseDuration.Duration,
-			RenewDeadline: vca.LeaderElection.RenewDeadline.Duration,
-			RetryPeriod:   vca.LeaderElection.RetryPeriod.Duration,
-			Callbacks: leaderelection.LeaderCallbacks{
-				OnStartedLeading: vca.onStartedLeading(),
-				OnStoppedLeading: func() {
-					golog.Fatal("leaderelection lost")
-				},
-			},
-		})
-	if err != nil {
+	if err := vca.setupLeaderElector(); err != nil {
 		golog.Fatal(err)
 	}
 
 	readyGauge.Set(1)
-	leaderElector.Run(vca.ctx)
+	vca.leaderElector.Run(vca.ctx)
 	readyGauge.Set(0)
 	panic("unreachable")
 }
@@ -385,22 +447,29 @@ func (vca *VirtControllerApp) onStartedLeading() func(ctx context.Context) {
 			vca.vmControllerThreads, vca.migrationControllerThreads, vca.evacuationControllerThreads,
 			vca.disruptionBudgetControllerThreads)
 
+		vmiprom.SetupVMICollector(vca.vmiInformer)
+		perfscale.RegisterPerfScaleMetrics(vca.vmiInformer)
+
 		go vca.evacuationController.Run(vca.evacuationControllerThreads, stop)
 		go vca.disruptionBudgetController.Run(vca.disruptionBudgetControllerThreads, stop)
 		go vca.nodeController.Run(vca.nodeControllerThreads, stop)
 		go vca.vmiController.Run(vca.vmiControllerThreads, stop)
 		go vca.rsController.Run(vca.rsControllerThreads, stop)
+		go vca.poolController.Run(vca.poolControllerThreads, stop)
 		go vca.vmController.Run(vca.vmControllerThreads, stop)
 		go vca.migrationController.Run(vca.migrationControllerThreads, stop)
 		go vca.snapshotController.Run(vca.snapshotControllerThreads, stop)
 		go vca.restoreController.Run(vca.restoreControllerThreads, stop)
+		go vca.workloadUpdateController.Run(stop)
+		go vca.nodeTopologyUpdater.Run(vca.nodeTopologyUpdatePeriod, stop)
+
 		cache.WaitForCacheSync(stop, vca.persistentVolumeClaimInformer.HasSynced)
 		close(vca.readyChan)
 		leaderGauge.Set(1)
 	}
 }
 
-func (vca *VirtControllerApp) getNewRecorder(namespace string, componentName string) record.EventRecorder {
+func (vca *VirtControllerApp) newRecorder(namespace string, componentName string) record.EventRecorder {
 	eventBroadcaster := record.NewBroadcaster()
 	eventBroadcaster.StartRecordingToSink(&k8coresv1.EventSinkImpl{Interface: vca.clientSet.CoreV1().Events(namespace)})
 	return eventBroadcaster.NewRecorder(scheme.Scheme, k8sv1.EventSource{Component: componentName})
@@ -416,6 +485,7 @@ func (vca *VirtControllerApp) initCommon() {
 
 	containerdisk.SetLocalDirectory(vca.ephemeralDiskDir + "/container-disk-data")
 	vca.templateService = services.NewTemplateService(vca.launcherImage,
+		vca.launcherQemuTimeout,
 		vca.virtShareDir,
 		vca.virtLibDir,
 		vca.ephemeralDiskDir,
@@ -428,42 +498,98 @@ func (vca *VirtControllerApp) initCommon() {
 		vca.launcherSubGid,
 	)
 
-	vca.vmiController = NewVMIController(vca.templateService, vca.vmiInformer, vca.kvPodInformer, vca.persistentVolumeClaimInformer, vca.vmiRecorder, vca.clientSet, vca.dataVolumeInformer)
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "node-controller")
+	topologyHinter := topology.NewTopologyHinter(vca.nodeInformer.GetStore(), vca.vmiInformer.GetStore(), runtime.GOARCH, vca.clusterConfig)
+
+	vca.vmiController = NewVMIController(vca.templateService,
+		vca.vmiInformer,
+		vca.vmInformer,
+		vca.kvPodInformer,
+		vca.persistentVolumeClaimInformer,
+		vca.vmiRecorder,
+		vca.clientSet,
+		vca.dataVolumeInformer,
+		vca.cdiInformer,
+		vca.cdiConfigInformer,
+		vca.clusterConfig,
+		topologyHinter,
+	)
+
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "node-controller")
 	vca.nodeController = NewNodeController(vca.clientSet, vca.nodeInformer, vca.vmiInformer, recorder)
-	vca.migrationController = NewMigrationController(vca.templateService, vca.vmiInformer, vca.kvPodInformer, vca.migrationInformer, vca.vmiRecorder, vca.clientSet, vca.clusterConfig)
+	vca.migrationController = NewMigrationController(
+		vca.templateService,
+		vca.vmiInformer,
+		vca.kvPodInformer,
+		vca.migrationInformer,
+		vca.nodeInformer,
+		vca.persistentVolumeClaimInformer,
+		vca.pdbInformer,
+		vca.migrationPolicyInformer,
+		vca.vmiRecorder,
+		vca.clientSet,
+		vca.clusterConfig,
+	)
+
+	vca.nodeTopologyUpdater = topology.NewNodeTopologyUpdater(vca.clientSet, topologyHinter, vca.nodeInformer)
 }
 
 func (vca *VirtControllerApp) initReplicaSet() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "virtualmachinereplicaset-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "virtualmachinereplicaset-controller")
 	vca.rsController = NewVMIReplicaSet(vca.vmiInformer, vca.rsInformer, recorder, vca.clientSet, controller.BurstReplicas)
 }
 
+func (vca *VirtControllerApp) initPool() {
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "virtualmachinepool-controller")
+	vca.poolController = NewPoolController(vca.clientSet,
+		vca.vmiInformer,
+		vca.vmInformer,
+		vca.poolInformer,
+		recorder,
+		controller.BurstReplicas)
+}
+
 func (vca *VirtControllerApp) initVirtualMachines() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "virtualmachine-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "virtualmachine-controller")
 
 	vca.vmController = NewVMController(
 		vca.vmiInformer,
 		vca.vmInformer,
 		vca.dataVolumeInformer,
 		vca.persistentVolumeClaimInformer,
+		vca.controllerRevisionInformer,
+		flavor.NewMethods(vca.flavorInformer.GetStore(), vca.clusterFlavorInformer.GetStore()),
 		recorder,
 		vca.clientSet)
 }
 
 func (vca *VirtControllerApp) initDisruptionBudgetController() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "disruptionbudget-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "disruptionbudget-controller")
 	vca.disruptionBudgetController = disruptionbudget.NewDisruptionBudgetController(
 		vca.vmiInformer,
-		vca.informerFactory.K8SInformerFactory().Policy().V1beta1().PodDisruptionBudgets().Informer(),
+		vca.pdbInformer,
+		vca.allPodInformer,
+		vca.migrationInformer,
 		recorder,
 		vca.clientSet,
 	)
 
 }
 
+func (vca *VirtControllerApp) initWorkloadUpdaterController() {
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "workload-update-controller")
+	vca.workloadUpdateController = workloadupdater.NewWorkloadUpdateController(
+		vca.launcherImage,
+		vca.vmiInformer,
+		vca.kvPodInformer,
+		vca.migrationInformer,
+		vca.kubeVirtInformer,
+		recorder,
+		vca.clientSet,
+		vca.clusterConfig)
+}
+
 func (vca *VirtControllerApp) initEvacuationController() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "disruptionbudget-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "disruptionbudget-controller")
 	vca.evacuationController = evacuation.NewEvacuationController(
 		vca.vmiInformer,
 		vca.migrationInformer,
@@ -476,7 +602,7 @@ func (vca *VirtControllerApp) initEvacuationController() {
 }
 
 func (vca *VirtControllerApp) initSnapshotController() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "snapshot-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "snapshot-controller")
 	vca.snapshotController = &snapshot.VMSnapshotController{
 		Client:                    vca.clientSet,
 		VMSnapshotInformer:        vca.vmSnapshotInformer,
@@ -488,6 +614,7 @@ func (vca *VirtControllerApp) initSnapshotController() {
 		CRDInformer:               vca.crdInformer,
 		PodInformer:               vca.allPodInformer,
 		DVInformer:                vca.dataVolumeInformer,
+		CRInformer:                vca.controllerRevisionInformer,
 		Recorder:                  recorder,
 		ResyncPeriod:              vca.snapshotControllerResyncPeriod,
 	}
@@ -495,7 +622,7 @@ func (vca *VirtControllerApp) initSnapshotController() {
 }
 
 func (vca *VirtControllerApp) initRestoreController() {
-	recorder := vca.getNewRecorder(k8sv1.NamespaceAll, "restore-controller")
+	recorder := vca.newRecorder(k8sv1.NamespaceAll, "restore-controller")
 	vca.restoreController = &snapshot.VMRestoreController{
 		Client:                    vca.clientSet,
 		VMRestoreInformer:         vca.vmRestoreInformer,
@@ -540,6 +667,9 @@ func (vca *VirtControllerApp) AddFlags() {
 	flag.StringVar(&vca.launcherImage, "launcher-image", launcherImage,
 		"Shim container for containerized VMIs")
 
+	flag.IntVar(&vca.launcherQemuTimeout, "launcher-qemu-timeout", launcherQemuTimeout,
+		"Amount of time to wait for qemu")
+
 	flag.StringVar(&vca.imagePullSecret, "image-pull-secret", imagePullSecret,
 		"Secret to use for pulling virt-launcher and/or registry disks")
 
@@ -562,12 +692,15 @@ func (vca *VirtControllerApp) AddFlags() {
 	flag.IntVar(&vca.nodeControllerThreads, "node-controller-threads", defaultControllerThreads,
 		"Number of goroutines to run for node controller")
 
-	flag.IntVar(&vca.vmiControllerThreads, "vmi-controller-threads", defaultControllerThreads,
+	flag.IntVar(&vca.vmiControllerThreads, "vmi-controller-threads", defaultVMIControllerThreads,
 		"Number of goroutines to run for vmi controller")
 
 	flag.IntVar(&vca.rsControllerThreads, "rs-controller-threads", defaultControllerThreads,
 		"Number of goroutines to run for replicaset controller")
 
+	flag.IntVar(&vca.poolControllerThreads, "pool-controller-threads", defaultControllerThreads,
+		"Number of goroutines to run for pool controller")
+
 	flag.IntVar(&vca.vmControllerThreads, "vm-controller-threads", defaultControllerThreads,
 		"Number of goroutines to run for vm controller")
 
@@ -583,7 +716,7 @@ func (vca *VirtControllerApp) AddFlags() {
 	flag.Int64Var(&vca.launcherSubGid, "launcher-subgid", defaultLauncherSubGid,
 		"ID of subgroup to virt-launcher")
 
-	flag.IntVar(&vca.snapshotControllerThreads, "snapshot-controller-threads", defaultControllerThreads,
+	flag.IntVar(&vca.snapshotControllerThreads, "snapshot-controller-threads", defaultSnapshotControllerThreads,
 		"Number of goroutines to run for snapshot controller")
 
 	flag.IntVar(&vca.restoreControllerThreads, "restore-controller-threads", defaultControllerThreads,
@@ -592,9 +725,59 @@ func (vca *VirtControllerApp) AddFlags() {
 	flag.DurationVar(&vca.snapshotControllerResyncPeriod, "snapshot-controller-resync-period", defaultSnapshotControllerResyncPeriod,
 		"Number of goroutines to run for snapshot controller")
 
+	flag.DurationVar(&vca.nodeTopologyUpdatePeriod, "node-topology-update-period", defaultNodeTopologyUpdatePeriod,
+		"Update period for the node topology updater")
+
 	flag.StringVar(&vca.promCertFilePath, "prom-cert-file", defaultPromCertFilePath,
 		"Client certificate used to prove the identity of the virt-controller when it must call out Promethus during a request")
 
 	flag.StringVar(&vca.promKeyFilePath, "prom-key-file", defaultPromKeyFilePath,
 		"Private key for the client certificate used to prove the identity of the virt-controller when it must call out Promethus during a request")
 }
+
+func (vca *VirtControllerApp) setupLeaderElector() (err error) {
+	clientConfig, err := kubecli.GetKubevirtClientConfig()
+	if err != nil {
+		return
+	}
+
+	clientConfig.RateLimiter =
+		flowcontrol.NewTokenBucketRateLimiter(
+			virtconfig.DefaultVirtControllerQPS,
+			virtconfig.DefaultVirtControllerBurst)
+
+	clientSet, err := kubecli.GetKubevirtClientFromRESTConfig(clientConfig)
+	if err != nil {
+		return
+	}
+
+	rl, err := resourcelock.New(vca.LeaderElection.ResourceLock,
+		vca.kubevirtNamespace,
+		leaderelectionconfig.DefaultEndpointName,
+		clientSet.CoreV1(),
+		clientSet.CoordinationV1(),
+		resourcelock.ResourceLockConfig{
+			Identity:      vca.host,
+			EventRecorder: vca.newRecorder(k8sv1.NamespaceAll, leaderelectionconfig.DefaultEndpointName),
+		})
+
+	if err != nil {
+		return
+	}
+
+	vca.leaderElector, err = leaderelection.NewLeaderElector(
+		leaderelection.LeaderElectionConfig{
+			Lock:          rl,
+			LeaseDuration: vca.LeaderElection.LeaseDuration.Duration,
+			RenewDeadline: vca.LeaderElection.RenewDeadline.Duration,
+			RetryPeriod:   vca.LeaderElection.RetryPeriod.Duration,
+			Callbacks: leaderelection.LeaderCallbacks{
+				OnStartedLeading: vca.onStartedLeading(),
+				OnStoppedLeading: func() {
+					golog.Fatal("leaderelection lost")
+				},
+			},
+		})
+
+	return
+}
